@@ -6,14 +6,18 @@ from starlette.exceptions import HTTPException as BaseHTTPException
 from starlette.types import ASGIApp, Message, Receive, Scope, Send
 
 from kupala.requests import Request
-from kupala.responses import HTMLResponse, Response
+from kupala.responses import GoBackResponse, HTMLResponse, JSONResponse, Response
 
 
-class HTTPException(BaseHTTPException):
+class KupalaError(Exception):
+    """Base class for all Kupala framework errors."""
+
+
+class HTTPException(BaseHTTPException, KupalaError):
     message: t.Optional[str] = None
 
     def __init__(self, status_code: int, message: str = None) -> None:  # pragma: nocover
-        message = message or self.message
+        self.message = message or self.message
         super().__init__(status_code=status_code, detail=message)
 
 
@@ -95,18 +99,52 @@ class TooManyRequests(_BasePredefinedHTTPException):
     status_code = 429
 
 
-ErrorHandler = t.Callable[[Request, Exception], t.Any]
+class ValidationError(BadRequest, KupalaError):
+    """Raised when data is not valid by any criteria."""
+
+    def __init__(
+        self,
+        message: str = None,
+        errors: t.Mapping[str, list[str]] = None,
+    ) -> None:
+        super().__init__(message)
+        self.message = message
+        self.errors = errors or {}
+
+
+E = t.TypeVar('E', bound=Exception)
+ErrorHandler = t.Callable[[Request, E], t.Any]
 
 _renderer = jinja2.Environment(loader=jinja2.PackageLoader(__name__.split('.')[0]))
 
 
-async def default_error_handler(request: Request, exc: HTTPException) -> Response:
-    content = _renderer.get_template('errors/http_error.html').render({'request': request, 'exc': exc})
-    return HTMLResponse(content, status_code=exc.status_code)
+async def default_validation_error_handler(request: Request, exc: ValidationError) -> Response:
+    if request.wants_json:
+        return JSONResponse({'message': exc.message, 'errors': exc.errors}, exc.status_code)
+    if 'session' in request.scope:
+        await request.remember_form_data()
+        request.set_form_errors(dict(exc.errors or {}), exc.message or '')
+    return GoBackResponse(request, exc.message, flash_category='error')
+
+
+async def default_http_error_handler(request: Request, exc: HTTPException) -> Response:
+    if request.wants_json:
+        data = {'message': exc.message, 'errors': getattr(exc, 'errors', {})}
+        if request.app.debug:
+            exception_type = f'{exc.__class__.__module__}.{exc.__class__.__qualname__}'
+            data['exception_type'] = exception_type
+            data['exception'] = repr(exc)
+        return JSONResponse(data, status_code=exc.status_code)
+    elif request.app.debug:
+        raise exc from None
+    else:
+        content = _renderer.get_template('errors/http_error.html').render({'request': request, 'exc': exc})
+        return HTMLResponse(content, status_code=exc.status_code)
 
 
 _default_error_handlers = {
-    BaseHTTPException: default_error_handler,
+    BaseHTTPException: default_http_error_handler,
+    ValidationError: default_validation_error_handler,
 }
 
 
@@ -116,9 +154,7 @@ class ErrorHandlers:
 
     def get_for_exception(self, exc: Exception) -> t.Optional[ErrorHandler]:
         class_name = type(exc)
-        status_code = 0
-        if isinstance(exc, BaseHTTPException):
-            status_code = exc.status_code
+        status_code = exc.status_code if isinstance(exc, BaseHTTPException) else 0
         if status_code in self.handlers:
             return self.handlers[status_code]
 
